package ru.dbotthepony.mc.otm.capability.android

import net.minecraft.MethodsReturnNonnullByDefault
import net.minecraft.core.Direction
import javax.annotation.ParametersAreNonnullByDefault
import net.minecraft.world.entity.LivingEntity
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraft.nbt.CompoundTag
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidFeature
import java.lang.Runnable
import net.minecraft.server.level.ServerPlayer
import ru.dbotthepony.mc.otm.network.android.AndroidFeaturePacket
import net.minecraftforge.event.entity.living.LivingHurtEvent
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.network.MatteryNetworking
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.network.android.AndroidEnergyPacket
import ru.dbotthepony.mc.otm.network.android.AndroidBatteryPacket
import net.minecraft.world.effect.MobEffect
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import net.minecraftforge.energy.CapabilityEnergy
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent
import net.minecraft.world.effect.MobEffects
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.eventbus.api.SubscribeEvent
import net.minecraftforge.event.entity.living.LivingEvent.LivingUpdateEvent
import net.minecraftforge.eventbus.api.EventPriority
import ru.dbotthepony.mc.otm.Registry
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.extractEnergy
import ru.dbotthepony.mc.otm.capability.receiveEnergy
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.set
import java.util.*

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
open class AndroidCapability(@JvmField protected val ent: LivingEntity) : ICapabilityProvider, IAndroidCapability, INBTSerializable<CompoundTag?> {
	@JvmField
	protected var energy_stored = Fraction.ZERO
	@JvmField
	protected var energy_stored_max = Fraction(60000)

	override var batteryItemStack = ItemStack.EMPTY

	private var network_energy = Fraction(-1)
	private var network_energy_max = Fraction(-1)
	private var network_battery = ItemStack.EMPTY

	@JvmField protected val features: MutableMap<AndroidFeatureType<*>?, AndroidFeature> = HashMap()
	@JvmField protected val network_queue = ArrayList<Any>()
	@JvmField protected val delayed_tick_server = ArrayList<Runnable>()
	@JvmField protected var network_first = false

	protected fun addFeature(feature: AndroidFeature): Boolean {
		if (features.containsKey(feature.type)) return false
		features[feature.type] = feature
		if (!ent.level.isClientSide) {
			delayed_tick_server.add(Runnable { feature.applyModifiers() })
		}
		if (ent is ServerPlayer) {
			sendNetwork(AndroidFeaturePacket(true, feature))
		}
		return true
	}

	override fun <T : AndroidFeature> addFeature(feature: AndroidFeatureType<T>): T {
		val get = features[feature]
		if (get != null) return get as T

		val factory = feature.factory(this)

		features[feature] = factory

		if (!ent.level.isClientSide) {
			delayed_tick_server.add(Runnable { factory!!.applyModifiers() })
		}

		if (ent is ServerPlayer) {
			sendNetwork(AndroidFeaturePacket(true, factory))
		}

		return factory
	}

	override fun removeFeature(feature: AndroidFeatureType<*>): Boolean {
		val removed = features.remove(feature)

		if (removed != null) {
			if (!ent.level.isClientSide) {
				delayed_tick_server.add { removed.removeModifiers() }
			}

			if (ent is ServerPlayer) {
				sendNetwork(AndroidFeaturePacket(false, removed))
			}

			return true
		}

		return false
	}

	override fun hasFeature(feature: AndroidFeatureType<*>): Boolean {
		return features.containsKey(feature)
	}

	override fun hasFeatureLevel(feature: AndroidFeatureType<*>, level: Int): Boolean {
		val get = features[feature] ?: return false
		return get.level >= level
	}

	override fun <T : AndroidFeature> getFeature(feature: AndroidFeatureType<T>): T? {
		return features[feature] as T?
	}

	override fun invalidateNetworkState() {
		network_energy = Fraction.MINUS_ONE
		network_energy_max = Fraction.MINUS_ONE
		network_battery = ItemStack.EMPTY

		if (ent is ServerPlayer) {
			for (feature in features.values) {
				sendNetwork(AndroidFeaturePacket(true, feature))
				feature.invalidateNetwork()
			}
		}
	}

	override fun setEnergy(value: Fraction?) {
		energy_stored = value!!
	}

	override fun setMaxEnergy(value: Fraction?) {
		energy_stored_max = value!!
	}

	override fun onHurt(event: LivingHurtEvent?) {
		for (feature in features.values) {
			feature.onHurt(event)
		}
	}

	override fun serializeNBT(): CompoundTag {
		val tag = CompoundTag()

		tag["energy_stored"] = energy_stored.serializeNBT()
		tag["energy_stored_max"] = energy_stored_max.serializeNBT()
		tag["battery"] = batteryItemStack.serializeNBT()

		val featureList = ListTag()

		for (feature in features.values) {
			val featureNbt = CompoundTag()
			feature.serializeNBT(featureNbt)

			featureNbt["id"] = feature.type.registryName!!.toString()
			featureList.add(featureNbt)
		}

		tag["features"] = featureList
		return tag
	}

	override fun deserializeNBT(compound: CompoundTag?) {
		compound!!

		compound.ifHas("energy_stored") {
			energy_stored = Fraction.deserializeNBT(it)
		}

		compound.ifHas("energy_stored_max") {
			energy_stored_max = Fraction.deserializeNBT(it)
		}

		compound.ifHas("battery", CompoundTag::class.java) {
			batteryItemStack = ItemStack.of(it)
		}

		features.clear()

		val featuresNbt = compound.getList("features", Tag.TAG_COMPOUND.toInt())

		for (tag in featuresNbt) {
			if (tag is CompoundTag) {
				val feature = Registry.ANDROID_FEATURES.getValue(ResourceLocation(tag.getString("id")))

				if (feature != null && feature.isApplicable(this)) {
					val feature = feature.factory(this)

					feature.deserializeNBT(tag)
					addFeature(feature)

					if (!ent.level.isClientSide) {
						delayed_tick_server.add(Runnable { feature.applyModifiers() })
					}
				}
			}
		}
	}

	fun dropBattery() {
		if (batteryItemStack.isEmpty) return
		ent.spawnAtLocation(batteryItemStack)
		batteryItemStack = ItemStack.EMPTY
	}

	protected fun sendNetwork(packet: Any) {
		if (ent is ServerPlayer) {
			if (network_first) {
				MatteryNetworking.CHANNEL.send(PacketDistributor.PLAYER.with { ent }, packet)
			} else {
				network_queue.add(packet)
			}
		}
	}

	protected open fun tickNetwork() {
		network_first = true

		if (ent is ServerPlayer) {
			if (energy_stored != network_energy) {
				network_energy = energy_stored
				sendNetwork(AndroidEnergyPacket(false, energy_stored))
			}

			if (energy_stored_max != network_energy_max) {
				network_energy_max = energy_stored_max
				sendNetwork(AndroidEnergyPacket(true, energy_stored_max))
			}

			if (!network_battery.equals(batteryItemStack, false)) {
				network_battery = batteryItemStack.copy()
				sendNetwork(AndroidBatteryPacket(batteryItemStack))
			}

			if (network_queue.size != 0) {
				for (packet in network_queue) {
					MatteryNetworking.CHANNEL.send(PacketDistributor.PLAYER.with {ent}, packet)
				}

				network_queue.clear()
			}
		}
	}

	protected open fun tickInnerClient() {}
	protected open fun tickInnerClientAlways() {}

	override fun tickClient() {
		delayed_tick_server.clear()
		if (!ent.isAlive) return
		tickInnerClientAlways()
		if (isAndroid()) {
			tickInnerClient()
			for (feature in features.values) {
				feature.tickClient()
			}
		}
	}

	override fun tick() {
		if (!ent.isAlive) return
		tickServerAlways()
		if (isAndroid()) {
			tickServer()
			for (feature in features.values) {
				feature.tickServer()
			}
		}
		for (runnable in delayed_tick_server) {
			runnable.run()
		}
		delayed_tick_server.clear()
		tickNetwork()
	}

	protected open fun tickServerAlways() {}

	protected open fun tickServer() {
		if (ent.airSupply < ent.maxAirSupply)
			ent.airSupply = ent.maxAirSupply

		for (effect in UNAFFECTED_EFFECTS)
			if (ent.hasEffect(effect))
				ent.removeEffect(effect)

		if (!batteryItemStack.isEmpty && energy_stored < energy_stored_max) {
			batteryItemStack.getCapability(CapabilityEnergy.ENERGY).ifPresent {
				if (it is IMatteryEnergyStorage) {
					energy_stored += it.extractEnergyInner(energy_stored_max - energy_stored, false)
				} else {
					energy_stored += it.extractEnergy(energy_stored_max - energy_stored, false)
				}
			}
		}
	}

	override fun extractEnergyOuter(howMuch: Fraction, simulate: Boolean): Fraction {
		return Fraction.ZERO
	}

	override fun extractEnergyInner(howMuch: Fraction, simulate: Boolean): Fraction {
		var howMuch = howMuch
		var drained = Fraction.ZERO

		if (!batteryItemStack.isEmpty) {
			batteryItemStack.getCapability(CapabilityEnergy.ENERGY).ifPresent {
				if (it is IMatteryEnergyStorage) {
					val extracted = it.extractEnergyOuter(howMuch, simulate)
					drained += extracted
					howMuch -= extracted
				} else {
					val extracted = it.extractEnergy(howMuch, simulate)
					drained += extracted
					howMuch -= extracted
				}
			}

			if (howMuch.isZero()) {
				if (!simulate && ent is ServerPlayer) {
					ent.awardStat(Registry.Names.POWER_CONSUMED, drained.toInt() * 10)
				}

				return drained
			}
		}

		val new = (energy_stored - howMuch).moreThanZero()
		drained += energy_stored - new

		if (!simulate) {
			energy_stored = new

			if (ent is ServerPlayer) {
				ent.awardStat(Registry.Names.POWER_CONSUMED, drained.toInt() * 10)
			}
		}

		return drained
	}

	override fun receiveEnergyOuter(howMuch: Fraction, simulate: Boolean): Fraction {
		var howMuch = howMuch
		var received = Fraction.ZERO

		if (!batteryItemStack.isEmpty) {
			batteryItemStack.getCapability(CapabilityEnergy.ENERGY).ifPresent {
				if (it is IMatteryEnergyStorage) {
					val extracted = it.receiveEnergyOuter(howMuch, simulate)
					received += extracted
					howMuch -= extracted
				} else {
					val extracted = it.receiveEnergy(howMuch, simulate)
					received += extracted
					howMuch -= extracted
				}
			}

			if (howMuch.isZero()) {
				return received
			}
		}

		val new = (energy_stored + howMuch).min(energy_stored_max)
		received += new - energy_stored

		if (!simulate) {
			energy_stored = new
		}

		return received
	}

	override fun receiveEnergyInner(howMuch: Fraction, simulate: Boolean): Fraction {
		return receiveEnergyOuter(howMuch, simulate)
	}

	override fun getEntity(): LivingEntity {
		return ent
	}

	override fun getBatteryLevel(): Fraction {
		if (!batteryItemStack.isEmpty) {
			val resolver = batteryItemStack.getCapability(CapabilityEnergy.ENERGY).resolve()

			if (resolver.isPresent) {
				val it = resolver.get()

				if (it is IMatteryEnergyStorage) {
					return energy_stored + it.batteryLevel
				} else {
					return energy_stored + it.energyStored
				}
			}
		}

		return energy_stored
	}

	override fun getMaxBatteryLevel(): Fraction {
		if (batteryItemStack != ItemStack.EMPTY) {
			val resolver = batteryItemStack.getCapability(CapabilityEnergy.ENERGY).resolve()

			if (resolver.isPresent) {
				val it = resolver.get()

				if (it is IMatteryEnergyStorage) {
					return energy_stored_max + it.maxBatteryLevel
				} else {
					return energy_stored_max + it.maxEnergyStored
				}
			}
		}

		return energy_stored_max
	}

	private val resolver = LazyOptional.of { this }

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (cap === MatteryCapability.ANDROID) {
			resolver.cast()
		} else LazyOptional.empty()
	}

	@Suppress("unused")
	companion object {
		@JvmField
		val UNAFFECTED_EFFECTS: MutableSet<MobEffect> = HashSet()

		@JvmStatic
		fun registerEffects(event: FMLCommonSetupEvent?) {
			UNAFFECTED_EFFECTS.add(MobEffects.CONDUIT_POWER)
			UNAFFECTED_EFFECTS.add(MobEffects.HEAL)
			// maybe it makes things go haywire idk
			// UNAFFECTED_EFFECTS.add(MobEffects.HARM);
			UNAFFECTED_EFFECTS.add(MobEffects.REGENERATION)
			UNAFFECTED_EFFECTS.add(MobEffects.WATER_BREATHING)
			UNAFFECTED_EFFECTS.add(MobEffects.POISON)
			// even skeletons can be withered
			// UNAFFECTED_EFFECTS.add(MobEffects.WITHER);
			UNAFFECTED_EFFECTS.add(MobEffects.HEALTH_BOOST)
			UNAFFECTED_EFFECTS.add(MobEffects.ABSORPTION)
			UNAFFECTED_EFFECTS.add(MobEffects.SATURATION)
			UNAFFECTED_EFFECTS.add(MobEffects.DOLPHINS_GRACE)
		}

		@SubscribeEvent
		fun onLivingTick(event: LivingUpdateEvent) {
			val ent = event.entity

			if (ent.level.isClientSide) {
				ent.getCapability(MatteryCapability.ANDROID).ifPresent {
					it.tickClient()
				}
			} else {
				ent.getCapability(MatteryCapability.ANDROID).ifPresent {
					it.tick()
				}
			}
		}

		@SubscribeEvent(priority = EventPriority.LOWEST)
		fun onHurtEvent(event: LivingHurtEvent) {
			if (event.isCanceled) return

			val ent = event.entity
			ent.getCapability(MatteryCapability.ANDROID).ifPresent { cap: IAndroidCapability -> cap.onHurt(event) }
		}
	}
}